
import asyncio
import datetime


def dprint(message_type: str, *args) -> None:
	print(str(datetime.datetime.now()) + ": (" + asyncio.current_task().get_name() + ") " + message_type + ": \n\t", *args)



import asyncio
import unittest

from aiosm import Host, Client, __version__


class Adder(Client):  # example implementing the Client class
	def __init__(self, name):
		super().__init__(name)
		self.white_list_functions += [  # names of class functions appended to the list are accessible by the Host
			"echo",
			"print",
			"pass_along",
			"add"
		]

		self.i = 1

	async def echo(self, message):
		await self.broadcast("test", "print", message)

	@staticmethod
	async def print(message):
		print(message)

	async def pass_along(self, tag):
		await self.broadcast(tag, "add", self.i)

	async def add(self, n):
		self.i += n


class MyTestCase(unittest.TestCase):
	def test_version(self):
		self.assertEqual(__version__, '0.0.5')

	host = Host()
	client1 = Adder("client1")
	client2 = Adder("client2")

	async def run_clients(self, callback):
		async def setup():
			await asyncio.gather(
				self.client1.connect(),
				self.client2.connect()
			)
			await asyncio.gather(  # subscription should occur after connecting yet before any broadcasting in callback
				self.client1.request("subscribe", "client1"),
				self.client2.request("subscribe", "client2")
			)
			await asyncio.gather(
				self.client1.run(),
				self.client2.run(),
				callback()
			)
		await self.host.run()
		await setup()

	def test_bidirectional_rpc(self):
		async def broadcast_test():
			asyncio.current_task().set_name("Broadcast Test")
			await self.client1.broadcast("client2", "pass_along", "client1")
			await self.client1.pass_along("client2")
			await self.client2.pass_along("client1")

			await asyncio.sleep(1)

			await self.host.close()

		asyncio.run(self.run_clients(broadcast_test))

		self.assertEqual(3, self.client1.i)
		self.assertEqual(2, self.client2.i)


if __name__ == '__main__':
	unittest.main()




from .radio import Radio
from .dprint import dprint

import asyncio
import inspect
import time
import json
from typing import List, Callable


def json_encoder(obj):
	if isinstance(obj, set):
		return list(obj)
	return obj.list()


class Responder(Radio):
	def __init__(self, name, reader=None, writer=None):
		super().__init__(reader, writer)

		self.__name__ = type(self).__name__ + "-" + name

		self.white_list_functions: List[str] = [  # todo change to function reference in a json friendly way
			"echo",
			"close"
		]

		self.clock = time.time()
		self.update_rate = 1/60

		self.ready = reader is not None

	async def callback(self, response: dict) -> None:
		if "type" in response and response["type"] in self.white_list_functions:
			function: Callable = getattr(self, response["type"])
			if "args" in response and response["args"] is not None:
				args = response["args"]
			else:
				args = ()

			if inspect.iscoroutinefunction(function):
				await function(*args)
			else:
				function(*args)
		else:
			print("(", asyncio.current_task().get_name(), ") Request unrecognised by server: " + str(response))

	async def run(self) -> None:
		asyncio.current_task().set_name(self.__name__ + "-Receiver")
		dprint("listening", "Begun")
		if self.ready:
			try:
				while self.ready:
					message = await self.receive()
					await self.callback(message)
			except ConnectionResetError:
				dprint("Error", "Connection lost")
			finally:
				self.close()
				if self.writer.close():
					await self.writer.wait_closed()
					# self.reader, self.writer = None, None
				dprint("listening", "Ended")

	async def wait(self):
		new_clock = time.time()
		await asyncio.sleep(self.update_rate + self.clock - new_clock)
		self.clock = new_clock

	@staticmethod
	def echo(message: str) -> None:
		print(message)

	def close(self) -> None:
		self.ready = False

	def prepare(self, json_dict: dict) -> bytes:
		message = json.dumps(json_dict, default=json_encoder)
		return super().prepare(message)

	def unpack(self, data: bytes) -> dict:
		message = super().unpack(data)
		return json.loads(message)

	def sequence(self, json_dict: dict) -> None:
		if not self.ready:
			raise Exception("Illegal communication")
		# noinspection PyTypeChecker
		super().sequence(json_dict)

	async def send(self, json_dict: dict) -> None:
		if not self.ready:
			raise Exception("Illegal communication")
		# noinspection PyTypeChecker
		await super().send(json_dict)

	async def feedback(self, json_dict: dict) -> None:
		if not self.ready:
			raise Exception("Illegal communication")
		# noinspection PyTypeChecker
		await super().feedback(json_dict)

	async def receive(self) -> dict:
		if not self.ready:
			raise Exception("Illegal communication")
		# noinspection PyTypeChecker
		return await super().receive()

	async def quit(self) -> None:
		json_dict = {"type": "close"}
		await self.send(json_dict)
		await self.feedback(json_dict)




from .dprint import dprint

from asyncio import StreamReader, StreamWriter


class Radio:
	ETX = b'\x03'  # end of transmission tag
	ETX_LEN = len(ETX)

	def __init__(self, reader: StreamReader, writer: StreamWriter):
		self.reader = reader
		self.writer = writer

	def prepare(self, message: str) -> bytes:
		# todo: use header
		data = message.encode()
		if self.ETX in data:
			raise Exception('message contains exit sequence: ' + self.ETX.decode())  # todo create custom Exception
		return data + self.ETX

	def unpack(self, data: bytes) -> str:
		# todo: use header
		return data[0:-self.ETX_LEN].decode()

	def sequence(self, message: str) -> None:
		data = self.prepare(message)
		self.writer.write(data)

	async def send(self, message: str) -> None:
		dprint("Sending", message)
		self.sequence(message)
		await self.writer.drain()

	async def feedback(self, message: str) -> None:
		data = self.prepare(message)
		dprint("Feeding-back", message)
		self.reader.feed_data(data)

	async def receive(self) -> str:
		data = await self.reader.readuntil(self.ETX)
		message = self.unpack(data)
		dprint("Received", message)
		return message



from .host import Host

import asyncio
import argparse

parser = argparse.ArgumentParser(description='Asynchronous socket manager for using asyncio with RPC capabilities.')
parser.add_argument('--addr', type=str, nargs='?', default='127.0.0.1', help='The IP address to use')
parser.add_argument('--port', type=int, nargs='?', default=8888, help='The port to use')
args = parser.parse_args()


def run_host():
	host = Host(addr=args.addr, port=args.port)
	loop = asyncio.get_event_loop()
	task = loop.create_task(host.run())
	task.set_name("Host")
	loop.run_forever()
	loop.close()


if __name__ == "__main__":
	run_host()




from .responder import Responder
# from .host import Host  # fixme circular import preventing type hinting

from typing import Set


class Node(Responder):
	def __init__(self, name, host, reader, writer):
		super().__init__(name, reader, writer)
		self.host = host

		self.subscriptions: Set[str] = {"all"}
		self.white_list_functions += [
			"subscribe",
			"broadcast"
		]

	async def subscribe(self, tag: str):
		self.subscriptions.add(tag)

	async def broadcast(self, json_dict: dict, tag: str):
		for node in self.host.connections:
			if node.ready and tag in node.subscriptions:  # todo change to channels containing Nodes to avoid for loops
				await node.send(json_dict)




from .dprint import dprint
from .node import Node

import asyncio
from asyncio import StreamReader, StreamWriter, AbstractServer
from typing import List


class Host:
	server: AbstractServer

	def __init__(self, addr='127.0.0.1', port=8888):
		super().__init__()
		self.addr = addr
		self.port = port
		self.connections: List[Node] = list()

	async def handler(self, reader: StreamReader, writer: StreamWriter):
		name = asyncio.current_task().get_name()
		node = Node(name, self, reader, writer)
		self.connections.append(node)
		try:
			await node.run()
		finally:
			self.connections.remove(node)

	async def run(self):
		self.server = await asyncio.start_server(self.handler, self.addr, self.port)
		await self.server.start_serving()
		dprint("Starting", f'Serving on {self.server.sockets[0].getsockname()}')

	async def close(self):
		for node in self.connections:
			await node.quit()
		self.server.close()
		await self.server.wait_closed()
		print("moo")



"""An asynchronous network hub

todo:
    revise Radio to use custom sockets
        create Radio using asyncio.create_datagram_endpoint() (aka UDP)
    revert to using headers
        use EOF marked for header?
    make encoder/decoder modules
        make protobuf encoder/decoders
    revise node communication
        replace JSON reliance with encoder/decoder module support
            perhaps an intermediate higher level interface for utility functions
    retest with JazZy
    implement heartbeat
        checking if a connection is still alive
    remember what old notes meant:
        implement inheritable class structure for client side to represent a node with header functions for RPC

"""

__version__ = '0.0.5'

from .__main__ import run_host
from .radio import Radio
from .clinet import Client
from .host import Host
from .node import Node
from .responder import Responder





from .responder import Responder

import asyncio


class Client(Responder):
	def __init__(self, name, addr='127.0.0.1', port=8888):
		super().__init__(name)
		self.addr = addr
		self.port = port

		# todo consider adding white_list_functions
		# self.white_list_functions += []

	async def connect(self) -> None:
		# todo add exception handling
		self.reader, self.writer = await asyncio.open_connection(self.addr, self.port)
		self.ready = True

	async def request(self, func_name: str, *args):
		await self.send({
			"type": func_name,
			"args": args
		})

	async def broadcast(self, tag: str, func_name: str, *args):  # only Clients subscribed to the tag shall receive
		# fixme, add key word args
		#  tdoo, remember what this meant
		await self.request("broadcast", {
					"type": func_name,
					"args": args
				},
				tag
		)
